'''
/*****************************************************************************************************
*                                   NOTICE
*
* THIS COMPUTER SOFTWARE CONTAINS PROPRIETARY INFORMATION OF USDA-ARS/ALARC.
* NEITHER RECEIPT NOR POSSESSION THEREOF COVERS ANY RIGHT TO
* REPRODUCE OR USE, OR DISCLOSE IN WHOLE OR IN PART, ANY SUCH INFORMATION
* WITHOUT WRITTEN AUTHORIZATION FROM USDA-ARS/ALARC.  
* 
* ****************************************************************************************************
* Program: RowDetection.py
* ****************************************************************************************************
* Purpose: Row detection using Hough Transform
* Author: James Kim (james.kim2@usda.gov)
* Version:  V1.0 2021-6-18  
* Revision:  
* ****************************************************************************************************/
'''
#lines = cv2.HoughLines(skel,r,th,threshold)
  # r: resol of r in pixel, default 1 pixel
  # th: resol of theta in radians, default 1 degree (CV_PI/180)
  # threshold: min # of intersections to detect a line
  # srn and stn: default parameters to 0
#lines = cv2.HoughLinesP(skel,r,th,threshold,0,minLineLength,maxLineGap)    
  # r: resol of r in pixel, default 1 pixel
  # th: resol of theta in radians, default 1 degree (CV_PI/180)
  # threshold: min # of intersections to detect a line
  # minLineLength: min # of points that can form a line 
  # maxLineGap: max gap between two points to be considered in the same line
# lines: vector that stores parameters (rho,theta) of the detected lines
  #if line is below origin, +rho & theta<180
  #if line is above origin, -rho & theta<180
  #if v-line, theta=0
  #if h-line, theta=90

path = '../Data/'

#import dependencies
import os, time
import tkinter as tk #cf)Tkinter for Python 2.*
from tkinter import filedialog,simpledialog,ttk
import cv2
import numpy as np
from math import *
import matplotlib.pyplot as plt
import imutils as imu   #rotate
import Stitch as stch    #stitch

#Load images in a directory
def load_images(src_dir):
    if not os.path.exists(src_dir):
        print('%s does not exist. Please enter a valid path.\n' % src_dir)

    Im = []
    Im_filename = []
    print('Loading images...')
    for filename in os.listdir(src_dir):
        if filename.endswith('.tif') or filename.endswith('.jpg'):
            #print('filename=',filename)
            im = cv2.imread(os.path.join(src_dir,filename))
            Im.append(im)
            Im_filename.append(filename)
    print('total # of images =',len(Im))
    return Im,Im_filename

# Geometric calibration 
def GeoCal_Image(src_dir,skew,stage):

    Im,Im_filename = load_images(src_dir)

    # deskew the images
    Im = stch.deskew_images(Im,skew)

    print('len(Im)=',len(Im))
    for i,im in enumerate (Im):
        
        # resize the image if too big
        scale = .5
        im = cv2.resize(im, (int(im.shape[1]*scale), int(im.shape[0]*scale))) 
        
        # Color shifting to greyscale
        #b,g,r = cv2.split(im)  # split the image into channels 
        #gscale = 2*g-r-b    # grayscale im by amplifing green to stand out 
        # growth stage [Early,Mid,Late] for row detection algorithm, e.g., winter wheat Early:Feb, Mid:March, Late:April
        if stage == 'Early':
            hsv = cv2.cvtColor(im,cv2.COLOR_RGB2HSV_FULL) #baseimage of imScn before segmentation
            h,s,gscale = cv2.split(hsv)  # split the image into channels         
            gscale = 255 - gscale # inverse the image to soil as foreground
            # segment to a binary image using Otsu's thresholding
            ret,mask = cv2.threshold(gscale,0,255,cv2.THRESH_BINARY+cv2.THRESH_OTSU)
        elif stage == 'Mid':
            hsv = cv2.cvtColor(im,cv2.COLOR_RGB2HSV_FULL) #baseimage of imScn before segmentation
            gscale,s,v = cv2.split(hsv)  # split the image into channels         
            mask = cv2.inRange(gscale,140,255)
        elif stage == 'Late':
            lab = cv2.cvtColor(im,cv2.COLOR_RGB2LAB) #baseimage of imScn before segmentation
            l,gscale,b = cv2.split(lab)  # split the image into channels                 
            # segment to a binary image using Otsu's thresholding
            ret,mask = cv2.threshold(gscale,0,255,cv2.THRESH_BINARY+cv2.THRESH_OTSU)
        
        #plt.subplot(221),plt.imshow(cv2.cvtColor(im,cv2.COLOR_BGR2RGB)),plt.title('im')
        #plt.subplot(222),plt.imshow(mask,cmap='gray'),plt.title('gscale')
        #plt.show()
        
        #mask = cv2.morphologyEx(mask,cv2.MORPH_OPEN,np.ones((2,2))) # Opening(erode+dilate) to fill frgnd holes
        mask = cv2.morphologyEx(mask,cv2.MORPH_CLOSE,np.ones((2,2))) # Closing(dilate+erode) to remove bkgnd particles 
        # Canny edge detection (adjust min,max)
        #edge = cv2.Canny(mask,0,50,apertureSize = 3)
        #edge = cv2.Canny(mask,400,400,apertureSize = 3)
        
        ## Skelectonize: thinning to binary
        #skel = np.zeros(edge.shape,np.uint8) 
        #element = cv2.getStructuringElement(cv2.MORPH_CROSS,(3,3))
        ##element = np.ones((2,2),np.uint8)
        #size = np.size(edge) # = edge.shape[0] x edge.shape[1]
        ##print('size=',size,'edge.shape=',edge.shape)
        #done = False
        #while( not done):
            #eroded = cv2.erode(edge,element)
            #temp = cv2.dilate(eroded,element)
            #temp = cv2.subtract(edge,temp)
            #skel = cv2.bitwise_or(skel,temp)
            #edge = eroded.copy()
            #zeros = size - cv2.countNonZero(edge)
            #print(size,cv2.countNonZero(edge))
            #if zeros==size:
                #done = True
        ctr_pt = [round(im.shape[0]/2),round(im.shape[1]/2)]
        Wp = 20   # search width % in the middle                 
        w = round(im.shape[1]*Wp/100/2)  # search width x offset
        ROI_x1,ROI_x2 = ctr_pt[1]-w, ctr_pt[1]+round(w/3)   # detect the left valley
        
        print('im.shape=',im.shape,'ctr_pt=',ctr_pt,'w=',w,'ROI_x1,ROI_x2=',ROI_x1,ROI_x2)
        if i==7: 
            if   stage == 'Early': ROI_y1 = round(mask.shape[0]*.5)  # trim the top area where the obstacles exist
            elif stage == 'Mid':   ROI_y1 = round(mask.shape[0]*.4)  # trim the top area where the obstacles exist
            elif stage == 'Late':  ROI_y1 = round(mask.shape[0]*.0)  # trim the top area where the obstacles exist
        else:    ROI_y1 = 0
        skel = mask[ROI_y1:,ROI_x1:ROI_x2] 
        
        # Line detection by Hough transform cv2.HoughLines(r,th,threshold,srn)
        #lines = cv2.HoughLines(skel,1,np.pi/180,130)
        #lines = cv2.HoughLinesP(skel,1,np.pi/180,150,0,skel.shape[0]*.9,30)    
        lines = cv2.HoughLinesP(skel,1,np.pi/180,150,0,skel.shape[0]*.5,60)    
          
        # draw the lines detected
        try:
            ImH,ImW = im.shape[:2]
            a,b,c = lines.shape
            print('lines.size=',lines.size,'lines.shape=',lines.shape,lines[0],lines[0][0])
            #for j in range(a):
            for j in range(1):
                #rho,theta = lines[j][0][:2]
                #a,b = cos(theta), sin(theta)
                #x0,y0 = a*rho, b*rho
                #print(a,b,x0,y0)
                #x1 = int(x0 + 1000*(-b))
                #y1 = int(y0 + 1000*(a))
                #x2 = int(x0 - 1000*(-b))
                #y2 = int(y0 - 1000*(a))    
                x1,y1,x2,y2 = lines[j][0]
                print('j=',j,'x1,y1,x2,y2=',x1,y1,x2,y2)
                cv2.line(im,(ROI_x1+x1,ROI_y1+y1),(ROI_x1+x2,ROI_y1+y2),(0,0,255),3,cv2.LINE_AA)
                cv2.line(mask,(ROI_x1+x1,ROI_y1+y1),(ROI_x1+x2,ROI_y1+y2),(0,0,255),3,cv2.LINE_AA)
            
            # calculate angle to rotate
            m = (y2-y1)/(x2-x1) # line eqn.:y=mx+b = m*x + y1-m*x1, b=y1-m*x1
            rot_deg = tanh(-1/m)*180/pi # rad -> deg
            print('m=',m,'rot_deg=tanh(-%d/%d)*180/pi=%f'%(x2-x1,y2-y1,rot_deg))
            
            # find extended x1,x2
            b = y2-m*x2
            y1,y2 = 0, ImH
            x1,x2 = round((y1-b)/m), round((y2-b)/m) # xx = (y-b)/m
            
            # rotate the image to align the line to vertially centered 
            im_rot = imu.rotate(im,rot_deg)
            #im_rot = imu.rotate_bound(im,rot_deg)
            
            ## rotate the image at an arbitrary point, not the center. (NOTE: same rotation with different trim area)
            #ctr_x = min(x1,x2)+round(abs(x1-x2)/2)
            #ctr_y = min(y1,y2)+round(abs(y1-y2)/2)
            #M = cv2.getRotationMatrix2D((ctr_x,ctr_y), rot_deg, 1.0)
            #im_rot = cv2.warpAffine(im, M, (im.shape[1], im.shape[0]))
            ##cv2.imshow("Rotated by Arbitrary Point", im_rot), cv2.waitKey()

            # shift the line to the center # NOTE: correct most but a few errors when short lines
            im_ctr_x = 300 #im_rot.shape[1]/2
            ctr_x = min(x1,x2)+round(abs(x1-x2)/2)
            move_x = im_ctr_x-ctr_x
            print('move_x=%d-%d=%d'%(im_ctr_x,ctr_x,move_x),'x1,y1,x2,y2=',x1,y1,x2,y2)
            # trim/add a margin
            im_rot_crop = np.zeros((ImH,ImW,3),np.uint8)
            #print('ImW,ImH=',ImW,ImH,'im_rot.shape',im_rot.shape,'im_rot_crop.shape=',im_rot_crop.shape)
            if move_x>0: im_rot_crop[:,move_x:] = im_rot[:,:ImW-move_x] # move the row to the right by adding left column space
            else:        im_rot_crop[:,:ImW-abs(move_x)] = im_rot[:,abs(move_x):] # move the row to the left by removing left column space
            #plt.subplot(221),plt.imshow(cv2.cvtColor(Im[i],cv2.COLOR_BGR2RGB)),plt.title('im')
            #plt.subplot(223),plt.imshow(cv2.cvtColor(im_rot,cv2.COLOR_BGR2RGB)),plt.title('im')
            #plt.subplot(224),plt.imshow(cv2.cvtColor(im_rot_crop,cv2.COLOR_BGR2RGB)),plt.title(move_x)
            #plt.show()
            im_rot_old = im_rot
            im_rot = im_rot_crop
        except: 
            im_rot = im

        #plt.subplot(221),plt.imshow(cv2.cvtColor(Im[i],cv2.COLOR_BGR2RGB)),plt.title('Input image')
        #plt.subplot(222),plt.imshow(gscale,cmap='gray'),plt.title('Grey scale')
        #plt.subplot(223),plt.imshow(cv2.cvtColor(im_rot,cv2.COLOR_BGR2RGB)),plt.title('Image rectified')
        #plt.subplot(224),plt.imshow(mask,cmap='gray'),plt.title('Binary')
        ##plt.subplot(224),plt.imshow(skel,cmap='gray'),plt.title('skel')

        plt.subplot(231),plt.imshow(cv2.cvtColor(Im[i],cv2.COLOR_BGR2RGB)),plt.title('Input image')
        plt.subplot(232),plt.imshow(gscale,cmap='gray'),plt.title('Grey scale')
        plt.subplot(233),plt.imshow(mask,cmap='gray'),plt.title('Binary')
        plt.subplot(234),plt.imshow(cv2.cvtColor(im_rot,cv2.COLOR_BGR2RGB)),plt.title('Image centered')
        plt.subplot(235),plt.imshow(cv2.cvtColor(im_rot_old,cv2.COLOR_BGR2RGB)),plt.title('Image rotated')
        plt.subplot(236),plt.imshow(skel,cmap='gray'),plt.title('Search area')
        plt.show()
                
        dst_dir = src_dir[:src_dir.find(src_dir.split('/')[-2])-1]+'_cal/'+src_dir.split('/')[-2]+'/'
        print(dst_dir)
        if not os.path.exists(dst_dir): os.makedirs(dst_dir)
        dst_file = dst_dir+Im_filename[i]
        print(i,dst_file)
        cv2.imwrite(dst_file,im_rot)
        
    return dst_dir

def main(src_dir,skew,stage):        
    # list subfolders
    subfolders = [ f.path for f in os.scandir(src_dir) if f.is_dir() ]
    print('src_dir=',src_dir,'subfolders=',subfolders)
    for i in range(len(subfolders)): subfolders[i] = subfolders[i].replace('\\','/')
    print('src_dir=',src_dir,'subfolders=',subfolders)
    if len(subfolders)>0:
        for i,im_dir in enumerate(subfolders):
            print(i,im_dir)
            dst_dir = GeoCal_Image(im_dir,skew,stage)
    else: 
        dst_dir = GeoCal_Image(src_dir,skew,stage)
    
    return dst_dir
    
if __name__ == '__main__':
    root = tk.Tk()
    root.title("Row Detection") # change the title of dialog widget        

    # load images from a folder
    src_dir = "C:/Users/James/Documents/IMAP/Data/DAQ2/Nadir/030821/A-1-L/"
    filename = '030821_A1_L_(4).jpg'
    src_dir = tk.filedialog.askdirectory() 
    if len(src_dir)==0: exit() 
    
    start_time = time.time()
    main(src_dir,skew=30,stage='Mid')   
    print("Total program run time = %.3f seconds." % (time.time() - start_time))
